/***************************************************************************
                          aliens.c
  description: alien creation
                             -------------------
    begin                : Fri March 30 2001
    date                 : 20/07/2010
    copyright            : (C) 2001 Interlinux Ltd
    email                : dbrasher@interlinux.co.uk
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "structs.h" // global screen
#include "bombs.h"
#include "data.h"
#include "aliens.h"

#ifndef FRAMES_PER_SEC          // Should get via. structs .h, but just in case
#define FRAMES_PER_SEC  50      // not critical should approximate framerate.
#endif

#define ALIEN_EXPLODE_TIME    30

typedef struct _s_Alien {
	char            *name;      // Filename of the image.
	char            *hit_name;  // Filename of the hit_points.
	Sprite          *sprites;   // Array of alien Sprites
	Sprite          *hits;      // Array of hit_points Sprites
	unsigned int     max;       // Maximum number allocated.
	unsigned int     points;    // Points awarded for a hit.

	// The following have functions to change the defauts.
	unsigned int     count;     // number used, cannot exceed max.
	unsigned int     speed;     // number of pixels to move per iteration.
	unsigned int     odds;      // spawning odds
	unsigned int     killshots; // hits required to kill.
} Alien_s;


/* to add an alien, insert a new type in the 'aliens.h Alien_t enum' and
   insert a struct below in the same order */
static Alien_s Aliens[NUM_ALIENS] = {
	{ "alien_slow.gif", "hit10.gif", NULL, NULL, 30, 10,
	   10, 2, (1*FRAMES_PER_SEC), 1 },
	{ "alien_fast.gif", "hit20.gif", NULL, NULL, 20, 20,
	   0, 0, (1*FRAMES_PER_SEC), 1 },
	{ "alien_mega.gif", "hit40.gif", NULL, NULL, 10, 40,
	   0, 0, (1*FRAMES_PER_SEC), 4 },
};



/**
 *  aliens_init()
 *  allocates the buffers and loads the images.
 *
 *  Returns:                    -1 on error, otherwise count of Sprites.
**/
int aliens_init(void)
{
	int alien_type;
	int j, tally = 0;

	for (alien_type = SLOW_ALIEN; alien_type < NUM_ALIENS; ++alien_type)
	{
		Alien_s *alien = &Aliens[alien_type];
		if ((alien->sprites = malloc(alien->max * sizeof(Sprite))) == NULL)
			return -1;
		if ((alien->hits = malloc(alien->max * sizeof(Sprite))) == NULL)
			return -1;

		if ((alien->sprites[0].image = LoadImage2(alien->name, 1)) == NULL)
			return -1;
		++tally;
		for (j = 1; j < alien->max; ++j)
		{
			alien->sprites[j].image = alien->sprites[0].image;
			++tally;
		}
		if ((alien->hits[0].image = LoadImage2(alien->hit_name, 1)) == NULL)
			return -1;
		++tally;
		for (j = 1; j < alien->max; ++j)
		{
			alien->hits[j].image = alien->hits[0].image;
			++tally;
		}
	}
	return tally;
}


/**
 *  aliens_free()
 *  frees the buffers and images.
 *
**/
void aliens_free(void)
{
	int alien_type;

	for (alien_type = SLOW_ALIEN; alien_type < NUM_ALIENS; ++alien_type)
	{
		Alien_s *alien = &Aliens[alien_type];
		SDL_FreeSurface(alien->sprites[0].image);
		free(alien->sprites);
		SDL_FreeSurface(alien->hits[0].image);
		free(alien->hits);
	}
}


/**
 *  aliens_collide()
 *  checks all aliens for a collision with sprite, breaks on collision
 *
 *  @sprite:    which sprite
 *  @explode:   if non-zero, creates an explosion of ALIEN_EXPLODE_TIME timeout.
 *
 *  Returns:    non-zero on collision
**/
int aliens_collide(Sprite *sprite, int explode)
{
	int alien_type;
	int j;

	for (alien_type = SLOW_ALIEN; alien_type < NUM_ALIENS; ++alien_type)
	{
		for (j = 0; j < Aliens[alien_type].count; ++j)
		{
			Sprite *s = &Aliens[alien_type].sprites[j];
			if (s->alive && sprite_collide(s, sprite))
			{
				if (explode)
				{
					--s->hits;
					if (s->hits == 0)
					{
						Sprite *e = &Aliens[alien_type].hits[j];
						s->alive = 0;
						e->alive = ALIEN_EXPLODE_TIME;
						e->x = s->x;
						e->y = s->y;
					}
				}
				return Aliens[alien_type].points;
			}
		}
	}
	return 0;
}


/**
 *  aliens_set_count()
 *  set how many aliens to be used, clamped 0 to Aliens[alien_type].max.
 *
 *  @alien_type:    which alien type
 *  @count:         how many aliens desired.
 *
 *  Returns:        -1 on error, or the count actually set.
**/
int aliens_set_count(Alien_t alien_type, int count)
{
	if (alien_type < 0 || alien_type >= NUM_ALIENS) // sanity check
		return -1;
	if (count > Aliens[alien_type].max)
		count = Aliens[alien_type].max;  // must not be more than allocated
	Aliens[alien_type].count = count;
	return count;
}


/**
 *  aliens_set_speed()
 *  set how many pixels this alien_type is to move, clamped 1 to 16.
 *
 *  @alien_type:    which alien type
 *  @speed:         how many pixels for a move.
 *
 *  Returns:        -1 on error, or the speed actually set.
**/
int aliens_set_speed(Alien_t alien_type, int speed)
{
	if (alien_type < 0 || alien_type >= NUM_ALIENS) // sanity check
		return -1;
	if (speed < 0)
		speed = 1;              // default minimum, move 1 pixel
	if (speed > screen->w / 16)
		speed = screen->w / 16; // default maximum, move 16 pixels
	Aliens[alien_type].speed = speed;
	return speed;
}


/**
 *  aliens_set_odds()
 *  odds of spawning this alien_type, minimum clamped to 8.
 *
 *  @alien_type:    which alien type
 *  @count:         odds of spawning.
 *
 *  Returns:        -1 on error, or the odds actually set.
**/
int aliens_set_odds(Alien_t alien_type, int count)
{
	if (alien_type < 0 || alien_type >= NUM_ALIENS) // sanity check
		return -1;
	if (count < 8)
		count = 8;              // default minimum 8
	Aliens[alien_type].odds = count;
	return count;
}

#if 0 //Currently unused
/**
 *  aliens_set_shots()
 *  set how many hits to kill this alien_type, minimum clamped to 1.
 *
 *  @alien_type:    which alien type
 *  @count:         how many hits.
 *
 *  Returns:        -1 on error, or the killshots actually set.
**/
int aliens_set_shots(Alien_t alien_type, int count)
{
	if (alien_type < 0 || alien_type >= NUM_ALIENS) // sanity check
		return -1;
	if (count < 1)
		count = 1;              // default minimum 1
	Aliens[alien_type].killshots = count;
	return count;
}
#endif

/**
 *  aliens_add_bomb()
 *  drop a bomb from a random alien, FIXME: may not happen.
 *
**/
void aliens_add_bomb(void)
{
	Sprite *bomb;
	int alien_type;
	int i;

	//randomly choose an alien type for bombing
	alien_type = (rand()%NUM_ALIENS);
	if (alien_type == FAST_ALIEN)
		return;

	Alien_s *a = &Aliens[alien_type];;
	if (!a->count)
		return;

	i = (rand()%a->count);
	// FIXME: perhaps a loop here until alive alien found ??
	if (a->sprites[i].alive )
	{	//bomb starting point
		if ((bomb = bombs_add()) != NULL)
		{
			bomb->x = a->sprites[i].x +
				(a->sprites[i].image->w - bomb->image->w) / 2;
			bomb->y = a->sprites[i].y +
				a->sprites[i].image->h;
			bomb->alive = 1;
		}
	}
	return;
}


/**
 *  aliens_add()
 *  randomly add an alien to each alien_type.
 *
**/
void aliens_add(void)
{
	int alien_type;

	for (alien_type = NUM_ALIENS - 1; alien_type >= 0; --alien_type)
	{
		int j = Aliens[alien_type].count;

		if ((rand() % Aliens[alien_type].odds) != 0)
			continue;
		while (j)
		{
			--j;
			Sprite *s = &Aliens[alien_type].sprites[j];
			if (!s->alive)
			{
				int direction;
				//Set random travel direction
				do
					direction = (rand() % 3) - 1;
				while (direction == 0);
				s->step = direction * Aliens[alien_type].speed;
				//Set initial location
				if (s->step > 0)
					s->x = (screen->w - s->image->w) - 1;
				else
					s->x = 0;
				s->y = 20;

				//Make it live
				s->alive = 1;
				//Init the hits
				s->hits = Aliens[alien_type].killshots;
				break; // if we get here, we are done
			}
		}
	}
}


/**
 *  aliens_move()
 *  perform action on all aliens of all alien types.
 *
**/
void aliens_move(void)
{
	int alien_type, i;

	for (alien_type = NUM_ALIENS - 1; alien_type >= 0; --alien_type)
	{
		for (i = 0; i < Aliens[alien_type].count; ++i)
		{
			Sprite *s = &Aliens[alien_type].sprites[i];
			if (s->alive)
			{
				s->x += s->step;
				if (s->x < 0 || s->x >= screen->w - s->image->w)
				{
					s->step = -s->step;
					s->y += s->image->h;
					if (s->x < 0)
						s->x = 0;
					else
						s->x = (screen->w - s->image->w) - 1;
				}
			}
		}
	}
}


/**
 *  aliens_foreach()
 *  perform action on all aliens of all alien types.
 *
 *  @dofunc:        See SpriteAction enum in sprite.h
 *
**/
void aliens_foreach(SpriteFunc dofunc)
{
	int alien_type;

	for (alien_type = NUM_ALIENS - 1; alien_type >= 0; --alien_type)
	{
		Alien_s *a = &Aliens[alien_type];
		int j = a->max;

		while (j > 0)
		{
			--j;
			Sprite *s = &a->sprites[j];
			Sprite *e = &a->hits[j];

			if (s->alive)
				dofunc(s);
			if (e->alive)
			{
				if (dofunc == sprite_ops[SPRITE_ERASE])
					--e->alive;
				dofunc(e);
			}
		}
	}
}
